"""Pytest configuration and fixtures for Pulsim tests."""

import pytest
import sys
import os

# Add the build directory to path if running tests before installation
# This allows testing the compiled module directly
build_paths = [
    os.path.join(os.path.dirname(__file__), '..', '..', 'build', 'python'),
    os.path.join(os.path.dirname(__file__), '..', '..', 'build'),
]
for path in build_paths:
    abs_path = os.path.abspath(path)
    if os.path.exists(abs_path) and abs_path not in sys.path:
        sys.path.insert(0, abs_path)


@pytest.fixture
def simple_rc_circuit():
    """Create a simple RC circuit for testing."""
    import pulsim as sl

    circuit = sl.Circuit()
    circuit.add_voltage_source("V1", "in", "0", 5.0)
    circuit.add_resistor("R1", "in", "out", 1000.0)
    circuit.add_capacitor("C1", "out", "0", 1e-6, ic=0.0)
    return circuit


@pytest.fixture
def buck_converter_circuit():
    """Create a buck converter circuit for testing."""
    import pulsim as sl

    circuit = sl.Circuit()

    # Input voltage source
    circuit.add_voltage_source("Vdc", "vcc", "0", 48.0)

    # PWM control signal
    circuit.add_voltage_source("Vctrl", "ctrl", "0", 5.0)

    # High-side switch
    sw_params = sl.SwitchParams()
    sw_params.ron = 0.01
    sw_params.roff = 1e9
    sw_params.vth = 2.5
    circuit.add_switch("S1", "vcc", "sw", "ctrl", "0", sw_params)

    # Freewheeling diode
    diode_params = sl.DiodeParams()
    diode_params.ideal = True
    circuit.add_diode("D1", "0", "sw", diode_params)

    # LC filter
    circuit.add_inductor("L1", "sw", "out", 100e-6)
    circuit.add_capacitor("C1", "out", "0", 100e-6)

    # Load
    circuit.add_resistor("Rload", "out", "0", 10.0)

    return circuit


@pytest.fixture
def simulation_options():
    """Default simulation options for testing."""
    import pulsim as sl

    opts = sl.SimulationOptions()
    opts.tstart = 0.0
    opts.tstop = 1e-3
    opts.dt = 1e-6
    opts.dtmax = 10e-6
    opts.abstol = 1e-12
    opts.reltol = 1e-3
    opts.use_ic = True
    return opts



"""Tests for thermal simulation."""

import pytest
import pulsim as sl


class TestThermalModel:
    """Test thermal model construction."""

    def test_foster_network(self):
        """Test Foster network construction."""
        foster = sl.FosterNetwork()
        foster.stages = []

        stage1 = sl.ThermalRCStage()
        stage1.rth = 0.1
        stage1.cth = 0.01

        stage2 = sl.ThermalRCStage()
        stage2.rth = 0.2
        stage2.cth = 0.02

        foster.stages = [stage1, stage2]

        assert len(foster.stages) == 2
        assert foster.rth_total() == pytest.approx(0.3)  # 0.1 + 0.2

    def test_thermal_rc_stage_tau(self):
        """Test RC stage time constant."""
        stage = sl.ThermalRCStage()
        stage.rth = 0.5  # K/W
        stage.cth = 0.1  # J/K

        # tau = R * C = 0.5 * 0.1 = 0.05 s
        assert stage.tau() == pytest.approx(0.05)

    def test_thermal_model_construction(self):
        """Test thermal model construction."""
        model = sl.ThermalModel()
        model.device_name = "Q1"
        # Use Simple type to use rth_jc directly
        model.type = sl.ThermalNetworkType.Simple
        model.rth_jc = 0.5
        model.rth_cs = 0.3
        model.rth_sa = 1.0
        model.tj_max = 175.0
        model.tj_warn = 150.0

        assert model.device_name == "Q1"
        # rth_ja = rth_jc + rth_cs + rth_sa = 0.5 + 0.3 + 1.0 = 1.8
        assert model.rth_ja() == pytest.approx(1.8)


class TestThermalSimulator:
    """Test thermal simulator."""

    def test_thermal_simulator_basic(self):
        """Basic thermal simulation."""
        model = sl.create_mosfet_thermal("M1", rth_jc=0.5, rth_cs=0.3, rth_sa=1.0)

        thermal = sl.ThermalSimulator()
        thermal.add_model(model)
        thermal.set_ambient(25.0)
        thermal.initialize()

        assert thermal.ambient() == 25.0
        assert thermal.junction_temp("M1") == pytest.approx(25.0)

    def test_thermal_step(self):
        """Test thermal stepping with power input."""
        model = sl.create_mosfet_thermal("M1", rth_jc=0.5)

        thermal = sl.ThermalSimulator()
        thermal.add_model(model)
        thermal.set_ambient(25.0)
        thermal.initialize()

        # Apply 10W for 1ms
        for _ in range(100):
            thermal.step(1e-5, {"M1": 10.0})

        tj = thermal.junction_temp("M1")
        # Temperature should have risen
        assert tj > 25.0

    def test_temperature_adjustment(self):
        """Test Rds_on and Vth temperature adjustment."""
        thermal = sl.ThermalSimulator()

        # Rds_on increases with temperature
        rds_25 = 0.05  # 50mOhm at 25°C
        rds_100 = thermal.adjust_rds_on(rds_25, 100.0)
        assert rds_100 > rds_25

        # Vth decreases with temperature (negative TC)
        vth_25 = 3.0  # 3V at 25°C
        vth_100 = thermal.adjust_vth(vth_25, 100.0)
        assert vth_100 < vth_25


class TestFosterNetworkFitting:
    """Test Foster network curve fitting."""

    def test_fit_foster_network(self):
        """Fit Foster network from Zth curve."""
        # Typical Zth curve datasheet points (time, Zth)
        zth_curve = [
            (1e-5, 0.01),
            (1e-4, 0.05),
            (1e-3, 0.15),
            (1e-2, 0.35),
            (1e-1, 0.48),
            (1.0, 0.50),
        ]

        foster = sl.fit_foster_network(zth_curve, num_stages=4)

        assert len(foster.stages) == 4
        # Total Rth should be close to final Zth
        assert foster.rth_total() == pytest.approx(0.5, rel=0.2)

    def test_create_mosfet_thermal(self):
        """Test MOSFET thermal model creation."""
        model = sl.create_mosfet_thermal("Q1", rth_jc=0.8, rth_cs=0.5, rth_sa=2.0)

        assert model.device_name == "Q1"
        assert model.rth_jc == pytest.approx(0.8)
        assert model.rth_cs == pytest.approx(0.5)
        assert model.rth_sa == pytest.approx(2.0)
        assert model.type == sl.ThermalNetworkType.Foster
        assert len(model.foster.stages) > 0



#!/usr/bin/env python3
"""
GUI Integration Example for Pulsim

This example demonstrates how to use Pulsim's GUI integration features:
1. Component metadata for building palettes and property editors
2. Circuit validation with detailed diagnostics
3. Simulation control (pause/resume/stop)
4. Progress callbacks for progress bars
5. Schematic position storage for layout persistence
6. Streaming configuration for memory-efficient long simulations

Run this script to see examples of each feature in action.
"""

import pulsim
import threading
import time
import json


def example_component_metadata():
    """
    Example 1: Component Metadata for GUI Palettes

    Use ComponentRegistry to build component palettes organized by category
    and generate property editors with correct field types.
    """
    print("\n" + "=" * 60)
    print("Example 1: Component Metadata for GUI Palettes")
    print("=" * 60)

    registry = pulsim.ComponentRegistry.instance()

    # List all categories
    print("\nAvailable component categories:")
    for category in registry.all_categories():
        print(f"  - {category}")

    # Build a palette structure
    print("\nComponent palette structure:")
    for category in registry.all_categories():
        print(f"\n  [{category}]")
        for comp_type in registry.types_in_category(category):
            meta = registry.get(comp_type)
            print(f"    - {meta.display_name} ({meta.symbol_id})")

    # Show property editor fields for a resistor
    print("\n\nProperty editor for Resistor:")
    meta = registry.get(pulsim.ComponentType.Resistor)
    print(f"  Description: {meta.description}")
    print(f"  Pins:")
    for pin in meta.pins:
        print(f"    - {pin.name}: {pin.description}")
    print(f"  Parameters:")
    for param in meta.parameters:
        unit_str = f" [{param.unit}]" if param.unit else ""
        required_str = " (required)" if param.required else " (optional)"
        print(f"    - {param.display_name}{unit_str}{required_str}")
        print(f"      Type: {param.type.name}")
        if param.min_value is not None:
            print(f"      Min: {param.min_value}")
        if param.max_value is not None:
            print(f"      Max: {param.max_value}")


def example_circuit_validation():
    """
    Example 2: Circuit Validation with Detailed Diagnostics

    Use validate_circuit() to get structured diagnostics for GUI error display.
    """
    print("\n" + "=" * 60)
    print("Example 2: Circuit Validation with Detailed Diagnostics")
    print("=" * 60)

    # Create a circuit with errors and warnings
    circuit = pulsim.Circuit()

    # This circuit has issues:
    # - No ground reference
    # - Floating node
    circuit.add_resistor("R1", "a", "b", 1000.0)
    circuit.add_resistor("R2", "c", "d", 2000.0)  # Disconnected

    result = pulsim.validate_circuit(circuit)

    print(f"\nValidation result: {'VALID' if result.is_valid else 'INVALID'}")
    print(f"Has errors: {result.has_errors()}")
    print(f"Has warnings: {result.has_warnings()}")

    if result.has_errors():
        print("\nErrors (must be fixed):")
        for error in result.errors():
            print(f"  [{error.code.name}] {error.message}")
            if error.component_name:
                print(f"    Component: {error.component_name}")
            if error.node_name:
                print(f"    Node: {error.node_name}")
            # Get more details about the error code
            desc = pulsim.diagnostic_code_description(error.code)
            print(f"    Help: {desc}")

    if result.has_warnings():
        print("\nWarnings (may cause issues):")
        for warning in result.warnings():
            print(f"  [{warning.code.name}] {warning.message}")

    # Now create a valid circuit
    print("\n--- Creating a valid circuit ---")
    valid_circuit = pulsim.Circuit()
    valid_circuit.add_voltage_source("V1", "in", "0", 5.0)
    valid_circuit.add_resistor("R1", "in", "out", 1000.0)
    valid_circuit.add_resistor("R2", "out", "0", 1000.0)

    result = pulsim.validate_circuit(valid_circuit)
    print(f"Validation result: {'VALID' if result.is_valid else 'INVALID'}")


def example_simulation_control():
    """
    Example 3: Simulation Control (Pause/Resume/Stop)

    Use SimulationController for thread-safe control of long simulations.
    """
    print("\n" + "=" * 60)
    print("Example 3: Simulation Control (Pause/Resume/Stop)")
    print("=" * 60)

    # Create a circuit for longer simulation
    circuit = pulsim.Circuit()
    circuit.add_voltage_source("V1", "in", "0", 5.0)
    circuit.add_resistor("R1", "in", "out", 1000.0)
    circuit.add_capacitor("C1", "out", "0", 1e-6)  # 1uF

    # Create controller
    controller = pulsim.SimulationController()

    print(f"\nInitial state: {controller.state.name}")

    # Setup simulation options for longer run
    opts = pulsim.SimulationOptions()
    opts.tstart = 0.0
    opts.tstop = 0.01  # 10ms
    opts.dt = 1e-7     # 100ns timestep

    # Track progress updates
    progress_updates = []

    def progress_callback(progress):
        progress_updates.append(progress.progress_percent)
        # In a real GUI, you would update the progress bar here
        if len(progress_updates) % 10 == 0:
            print(f"  Progress: {progress.progress_percent:.1f}% "
                  f"(t={progress.current_time*1000:.3f}ms, "
                  f"steps={progress.steps_completed})")

    # Run simulation in a thread (like you would in a GUI)
    sim = pulsim.Simulator(circuit, opts)

    def run_simulation():
        sim.run_transient_with_progress(
            callback=None,
            event_callback=None,
            control=controller,
            progress_callback=progress_callback,
            min_interval_ms=50,  # Update every 50ms
            min_steps=100
        )

    print("\nStarting simulation in background thread...")
    sim_thread = threading.Thread(target=run_simulation)
    sim_thread.start()

    # Wait a bit, then pause
    time.sleep(0.1)
    if controller.is_running():
        print("\nRequesting pause...")
        controller.request_pause()
        time.sleep(0.05)
        print(f"State after pause request: {controller.state.name}")

    # Resume after a brief pause
    time.sleep(0.1)
    if controller.is_paused():
        print("\nRequesting resume...")
        controller.request_resume()

    # Wait for completion
    sim_thread.join(timeout=5.0)

    print(f"\nFinal state: {controller.state.name}")
    print(f"Total progress updates received: {len(progress_updates)}")


def example_progress_callbacks():
    """
    Example 4: Progress Callbacks for Progress Bars

    Configure progress callbacks to update GUI progress indicators.
    """
    print("\n" + "=" * 60)
    print("Example 4: Progress Callbacks for Progress Bars")
    print("=" * 60)

    circuit = pulsim.Circuit()
    circuit.add_voltage_source("V1", "in", "0", 5.0)
    circuit.add_resistor("R1", "in", "out", 1000.0)
    circuit.add_capacitor("C1", "out", "0", 1e-6)

    opts = pulsim.SimulationOptions()
    opts.tstart = 0.0
    opts.tstop = 0.005  # 5ms
    opts.dt = 1e-7

    # Configure progress callback throttling in options
    opts.progress_min_interval_ms = 100   # Max 10 updates/second
    opts.progress_min_steps = 50          # At least 50 steps between updates
    opts.progress_include_memory = False  # Don't track memory (faster)

    print("\nProgress callback configuration:")
    print(f"  Min interval: {opts.progress_min_interval_ms}ms")
    print(f"  Min steps: {opts.progress_min_steps}")
    print(f"  Track memory: {opts.progress_include_memory}")

    progress_data = []

    def detailed_progress_callback(p):
        progress_data.append({
            'percent': p.progress_percent,
            'time': p.current_time,
            'steps': p.steps_completed,
            'newton_iters': p.newton_iterations,
            'convergence_warning': p.convergence_warning,
            'elapsed': p.elapsed_seconds,
            'remaining': p.estimated_remaining_seconds
        })

    sim = pulsim.Simulator(circuit, opts)
    result = sim.run_transient_with_progress(
        callback=None,
        event_callback=None,
        control=None,
        progress_callback=detailed_progress_callback,
        min_interval_ms=100,
        min_steps=50
    )

    print(f"\nSimulation completed!")
    print(f"Total progress callbacks: {len(progress_data)}")

    if progress_data:
        print("\nSample progress data:")
        for i, p in enumerate(progress_data[::max(1, len(progress_data)//5)]):
            print(f"  {p['percent']:5.1f}% | t={p['time']*1000:.2f}ms | "
                  f"steps={p['steps']:6d} | elapsed={p['elapsed']:.3f}s")


def example_schematic_positions():
    """
    Example 5: Schematic Position Storage

    Store and retrieve component positions for GUI layout persistence.
    """
    print("\n" + "=" * 60)
    print("Example 5: Schematic Position Storage")
    print("=" * 60)

    # Create circuit and set positions
    circuit = pulsim.Circuit()
    circuit.add_voltage_source("V1", "in", "0", 5.0)
    circuit.add_resistor("R1", "in", "out", 1000.0)
    circuit.add_capacitor("C1", "out", "0", 1e-6)

    # Set positions (as if placed by user in schematic editor)
    circuit.set_position("V1", pulsim.SchematicPosition(x=50, y=100, orientation=0))
    circuit.set_position("R1", pulsim.SchematicPosition(x=150, y=100, orientation=0))
    circuit.set_position("C1", pulsim.SchematicPosition(x=250, y=150, orientation=90))

    print("\nComponent positions set:")
    for name in ["V1", "R1", "C1"]:
        pos = circuit.get_position(name)
        if pos:
            print(f"  {name}: x={pos.x}, y={pos.y}, rotation={pos.orientation}deg")

    # Export to JSON (includes positions)
    json_str = pulsim.circuit_to_json(circuit, include_positions=True)

    print("\nJSON export (with positions):")
    json_obj = json.loads(json_str)
    print(json.dumps(json_obj, indent=2)[:500] + "...")

    # Import from JSON
    loaded_circuit = pulsim.parse_netlist_string(json_str)

    print("\nPositions after import:")
    for name in ["V1", "R1", "C1"]:
        if loaded_circuit.has_position(name):
            pos = loaded_circuit.get_position(name)
            print(f"  {name}: x={pos.x}, y={pos.y}, rotation={pos.orientation}deg")
        else:
            print(f"  {name}: no position")


def example_streaming_configuration():
    """
    Example 6: Streaming Configuration for Long Simulations

    Configure memory-efficient result storage for long simulations.
    """
    print("\n" + "=" * 60)
    print("Example 6: Streaming Configuration for Long Simulations")
    print("=" * 60)

    circuit = pulsim.Circuit()
    circuit.add_voltage_source("V1", "in", "0", 5.0)
    circuit.add_resistor("R1", "in", "out", 1000.0)
    circuit.add_capacitor("C1", "out", "0", 1e-6)

    # Without streaming: many points
    opts_full = pulsim.SimulationOptions()
    opts_full.tstart = 0.0
    opts_full.tstop = 0.01  # 10ms
    opts_full.dt = 1e-7
    opts_full.streaming_decimation = 1  # Store all points

    sim_full = pulsim.Simulator(circuit, opts_full)
    result_full = sim_full.run_transient()

    print(f"\nWithout decimation:")
    print(f"  Points stored: {result_full.num_points()}")

    # With decimation: fewer points
    opts_decimated = pulsim.SimulationOptions()
    opts_decimated.tstart = 0.0
    opts_decimated.tstop = 0.01
    opts_decimated.dt = 1e-7
    opts_decimated.streaming_decimation = 100  # Store every 100th point

    sim_decimated = pulsim.Simulator(circuit, opts_decimated)
    result_decimated = sim_decimated.run_transient()

    print(f"\nWith decimation (every 100th point):")
    print(f"  Points stored: {result_decimated.num_points()}")

    # With rolling buffer
    opts_rolling = pulsim.SimulationOptions()
    opts_rolling.tstart = 0.0
    opts_rolling.tstop = 0.01
    opts_rolling.dt = 1e-7
    opts_rolling.streaming_rolling_buffer = True
    opts_rolling.streaming_max_points = 1000

    sim_rolling = pulsim.Simulator(circuit, opts_rolling)
    result_rolling = sim_rolling.run_transient()

    print(f"\nWith rolling buffer (max 1000 points):")
    print(f"  Points stored: {result_rolling.num_points()}")

    print("\nStreaming is useful for:")
    print("  - Long simulations where you only need recent data")
    print("  - Memory-constrained environments")
    print("  - Real-time displays that show a time window")


def example_enhanced_results():
    """
    Example 7: Enhanced Simulation Results

    Access detailed signal info, solver settings, and events from results.
    """
    print("\n" + "=" * 60)
    print("Example 7: Enhanced Simulation Results")
    print("=" * 60)

    circuit = pulsim.Circuit()
    circuit.add_voltage_source("V1", "in", "0", 5.0)
    circuit.add_resistor("R1", "in", "out", 1000.0)
    circuit.add_capacitor("C1", "out", "0", 1e-6)

    opts = pulsim.SimulationOptions()
    opts.tstart = 0.0
    opts.tstop = 0.005
    opts.dt = 1e-7
    opts.integration_method = pulsim.IntegrationMethod.Trapezoidal

    sim = pulsim.Simulator(circuit, opts)
    result = sim.run_transient()

    print("\nSignal information (for plot labels and axes):")
    for info in result.signal_info:
        print(f"  - {info.name}: type={info.type}, unit={info.unit}")

    print("\nSolver information:")
    print(f"  Method: {result.solver_info.method}")
    print(f"  Abstol: {result.solver_info.abstol}")
    print(f"  Reltol: {result.solver_info.reltol}")
    print(f"  Adaptive: {result.solver_info.adaptive_timestep}")

    print("\nPerformance metrics:")
    print(f"  Total steps: {result.total_steps}")
    print(f"  Avg Newton iterations: {result.average_newton_iterations:.2f}")
    print(f"  Convergence failures: {result.convergence_failures}")
    print(f"  Timestep reductions: {result.timestep_reductions}")

    print(f"\nEvents recorded: {result.num_events()}")
    if result.num_events() > 0:
        print("Sample events:")
        for event in result.events[:5]:
            print(f"  t={event.time:.6f}s: {event.description}")


def main():
    """Run all GUI integration examples."""
    print("=" * 60)
    print("Pulsim GUI Integration Examples")
    print("=" * 60)
    print("\nThis script demonstrates GUI integration features.")
    print("Each example shows how to use a specific feature.")

    example_component_metadata()
    example_circuit_validation()
    example_simulation_control()
    example_progress_callbacks()
    example_schematic_positions()
    example_streaming_configuration()
    example_enhanced_results()

    print("\n" + "=" * 60)
    print("All examples completed!")
    print("=" * 60)


if __name__ == "__main__":
    main()



import datetime

from google.protobuf import empty_pb2 as _empty_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf import duration_pb2 as _duration_pb2
from google.protobuf import wrappers_pb2 as _wrappers_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from collections.abc import Iterable as _Iterable, Mapping as _Mapping
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class HealthStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    HEALTH_STATUS_UNSPECIFIED: _ClassVar[HealthStatus]
    HEALTH_STATUS_OK: _ClassVar[HealthStatus]
    HEALTH_STATUS_DEGRADED: _ClassVar[HealthStatus]
    HEALTH_STATUS_ERROR: _ClassVar[HealthStatus]

class SessionStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    SESSION_STATUS_UNSPECIFIED: _ClassVar[SessionStatus]
    SESSION_STATUS_CREATED: _ClassVar[SessionStatus]
    SESSION_STATUS_VALIDATING: _ClassVar[SessionStatus]
    SESSION_STATUS_READY: _ClassVar[SessionStatus]
    SESSION_STATUS_RUNNING: _ClassVar[SessionStatus]
    SESSION_STATUS_PAUSED: _ClassVar[SessionStatus]
    SESSION_STATUS_COMPLETED: _ClassVar[SessionStatus]
    SESSION_STATUS_STOPPED: _ClassVar[SessionStatus]
    SESSION_STATUS_FAILED: _ClassVar[SessionStatus]

class SimulationResultFormat(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    SIMULATION_RESULT_FORMAT_UNSPECIFIED: _ClassVar[SimulationResultFormat]
    SIMULATION_RESULT_FORMAT_CSV: _ClassVar[SimulationResultFormat]
    SIMULATION_RESULT_FORMAT_HDF5: _ClassVar[SimulationResultFormat]
    SIMULATION_RESULT_FORMAT_PARQUET: _ClassVar[SimulationResultFormat]

class SweepStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    SWEEP_STATUS_UNSPECIFIED: _ClassVar[SweepStatus]
    SWEEP_STATUS_PENDING: _ClassVar[SweepStatus]
    SWEEP_STATUS_RUNNING: _ClassVar[SweepStatus]
    SWEEP_STATUS_COMPLETED: _ClassVar[SweepStatus]
    SWEEP_STATUS_FAILED: _ClassVar[SweepStatus]
    SWEEP_STATUS_CANCELLED: _ClassVar[SweepStatus]
HEALTH_STATUS_UNSPECIFIED: HealthStatus
HEALTH_STATUS_OK: HealthStatus
HEALTH_STATUS_DEGRADED: HealthStatus
HEALTH_STATUS_ERROR: HealthStatus
SESSION_STATUS_UNSPECIFIED: SessionStatus
SESSION_STATUS_CREATED: SessionStatus
SESSION_STATUS_VALIDATING: SessionStatus
SESSION_STATUS_READY: SessionStatus
SESSION_STATUS_RUNNING: SessionStatus
SESSION_STATUS_PAUSED: SessionStatus
SESSION_STATUS_COMPLETED: SessionStatus
SESSION_STATUS_STOPPED: SessionStatus
SESSION_STATUS_FAILED: SessionStatus
SIMULATION_RESULT_FORMAT_UNSPECIFIED: SimulationResultFormat
SIMULATION_RESULT_FORMAT_CSV: SimulationResultFormat
SIMULATION_RESULT_FORMAT_HDF5: SimulationResultFormat
SIMULATION_RESULT_FORMAT_PARQUET: SimulationResultFormat
SWEEP_STATUS_UNSPECIFIED: SweepStatus
SWEEP_STATUS_PENDING: SweepStatus
SWEEP_STATUS_RUNNING: SweepStatus
SWEEP_STATUS_COMPLETED: SweepStatus
SWEEP_STATUS_FAILED: SweepStatus
SWEEP_STATUS_CANCELLED: SweepStatus

class SimulationOptions(_message.Message):
    __slots__ = ("tstart", "tstop", "dt", "dtmin", "dtmax", "abstol", "reltol", "max_newton_iterations", "damping_factor", "use_ic", "output_signals")
    TSTART_FIELD_NUMBER: _ClassVar[int]
    TSTOP_FIELD_NUMBER: _ClassVar[int]
    DT_FIELD_NUMBER: _ClassVar[int]
    DTMIN_FIELD_NUMBER: _ClassVar[int]
    DTMAX_FIELD_NUMBER: _ClassVar[int]
    ABSTOL_FIELD_NUMBER: _ClassVar[int]
    RELTOL_FIELD_NUMBER: _ClassVar[int]
    MAX_NEWTON_ITERATIONS_FIELD_NUMBER: _ClassVar[int]
    DAMPING_FACTOR_FIELD_NUMBER: _ClassVar[int]
    USE_IC_FIELD_NUMBER: _ClassVar[int]
    OUTPUT_SIGNALS_FIELD_NUMBER: _ClassVar[int]
    tstart: _wrappers_pb2.DoubleValue
    tstop: _wrappers_pb2.DoubleValue
    dt: _wrappers_pb2.DoubleValue
    dtmin: _wrappers_pb2.DoubleValue
    dtmax: _wrappers_pb2.DoubleValue
    abstol: _wrappers_pb2.DoubleValue
    reltol: _wrappers_pb2.DoubleValue
    max_newton_iterations: _wrappers_pb2.Int32Value
    damping_factor: _wrappers_pb2.DoubleValue
    use_ic: _wrappers_pb2.BoolValue
    output_signals: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, tstart: _Optional[_Union[_wrappers_pb2.DoubleValue, _Mapping]] = ..., tstop: _Optional[_Union[_wrappers_pb2.DoubleValue, _Mapping]] = ..., dt: _Optional[_Union[_wrappers_pb2.DoubleValue, _Mapping]] = ..., dtmin: _Optional[_Union[_wrappers_pb2.DoubleValue, _Mapping]] = ..., dtmax: _Optional[_Union[_wrappers_pb2.DoubleValue, _Mapping]] = ..., abstol: _Optional[_Union[_wrappers_pb2.DoubleValue, _Mapping]] = ..., reltol: _Optional[_Union[_wrappers_pb2.DoubleValue, _Mapping]] = ..., max_newton_iterations: _Optional[_Union[_wrappers_pb2.Int32Value, _Mapping]] = ..., damping_factor: _Optional[_Union[_wrappers_pb2.DoubleValue, _Mapping]] = ..., use_ic: _Optional[_Union[_wrappers_pb2.BoolValue, _Mapping]] = ..., output_signals: _Optional[_Iterable[str]] = ...) -> None: ...

class CircuitModel(_message.Message):
    __slots__ = ("name", "description", "model_json", "created_at", "created_by")
    NAME_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    MODEL_JSON_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    CREATED_BY_FIELD_NUMBER: _ClassVar[int]
    name: str
    description: str
    model_json: str
    created_at: _timestamp_pb2.Timestamp
    created_by: str
    def __init__(self, name: _Optional[str] = ..., description: _Optional[str] = ..., model_json: _Optional[str] = ..., created_at: _Optional[_Union[datetime.datetime, _timestamp_pb2.Timestamp, _Mapping]] = ..., created_by: _Optional[str] = ...) -> None: ...

class ModelDescriptor(_message.Message):
    __slots__ = ("model_id", "model")
    MODEL_ID_FIELD_NUMBER: _ClassVar[int]
    MODEL_FIELD_NUMBER: _ClassVar[int]
    model_id: str
    model: CircuitModel
    def __init__(self, model_id: _Optional[str] = ..., model: _Optional[_Union[CircuitModel, _Mapping]] = ...) -> None: ...

class SessionDescriptor(_message.Message):
    __slots__ = ("session_id", "model_id", "name", "status", "created_at", "updated_at", "active_signals", "retention", "owner")
    SESSION_ID_FIELD_NUMBER: _ClassVar[int]
    MODEL_ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    UPDATED_AT_FIELD_NUMBER: _ClassVar[int]
    ACTIVE_SIGNALS_FIELD_NUMBER: _ClassVar[int]
    RETENTION_FIELD_NUMBER: _ClassVar[int]
    OWNER_FIELD_NUMBER: _ClassVar[int]
    session_id: str
    model_id: str
    name: str
    status: SessionStatus
    created_at: _timestamp_pb2.Timestamp
    updated_at: _timestamp_pb2.Timestamp
    active_signals: _containers.RepeatedScalarFieldContainer[str]
    retention: _duration_pb2.Duration
    owner: str
    def __init__(self, session_id: _Optional[str] = ..., model_id: _Optional[str] = ..., name: _Optional[str] = ..., status: _Optional[_Union[SessionStatus, str]] = ..., created_at: _Optional[_Union[datetime.datetime, _timestamp_pb2.Timestamp, _Mapping]] = ..., updated_at: _Optional[_Union[datetime.datetime, _timestamp_pb2.Timestamp, _Mapping]] = ..., active_signals: _Optional[_Iterable[str]] = ..., retention: _Optional[_Union[datetime.timedelta, _duration_pb2.Duration, _Mapping]] = ..., owner: _Optional[str] = ...) -> None: ...

class HealthCheckRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class HealthCheckResponse(_message.Message):
    __slots__ = ("status", "version", "uptime", "active_sessions", "completed_sessions", "authentication_enabled")
    STATUS_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    UPTIME_FIELD_NUMBER: _ClassVar[int]
    ACTIVE_SESSIONS_FIELD_NUMBER: _ClassVar[int]
    COMPLETED_SESSIONS_FIELD_NUMBER: _ClassVar[int]
    AUTHENTICATION_ENABLED_FIELD_NUMBER: _ClassVar[int]
    status: HealthStatus
    version: str
    uptime: _duration_pb2.Duration
    active_sessions: int
    completed_sessions: int
    authentication_enabled: bool
    def __init__(self, status: _Optional[_Union[HealthStatus, str]] = ..., version: _Optional[str] = ..., uptime: _Optional[_Union[datetime.timedelta, _duration_pb2.Duration, _Mapping]] = ..., active_sessions: _Optional[int] = ..., completed_sessions: _Optional[int] = ..., authentication_enabled: bool = ...) -> None: ...

class CreateSessionRequest(_message.Message):
    __slots__ = ("name", "model_id", "inline_model", "options")
    NAME_FIELD_NUMBER: _ClassVar[int]
    MODEL_ID_FIELD_NUMBER: _ClassVar[int]
    INLINE_MODEL_FIELD_NUMBER: _ClassVar[int]
    OPTIONS_FIELD_NUMBER: _ClassVar[int]
    name: str
    model_id: str
    inline_model: CircuitModel
    options: SimulationOptions
    def __init__(self, name: _Optional[str] = ..., model_id: _Optional[str] = ..., inline_model: _Optional[_Union[CircuitModel, _Mapping]] = ..., options: _Optional[_Union[SimulationOptions, _Mapping]] = ...) -> None: ...

class CreateSessionResponse(_message.Message):
    __slots__ = ("session",)
    SESSION_FIELD_NUMBER: _ClassVar[int]
    session: SessionDescriptor
    def __init__(self, session: _Optional[_Union[SessionDescriptor, _Mapping]] = ...) -> None: ...

class ListSessionsRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class ListSessionsResponse(_message.Message):
    __slots__ = ("sessions",)
    SESSIONS_FIELD_NUMBER: _ClassVar[int]
    sessions: _containers.RepeatedCompositeFieldContainer[SessionDescriptor]
    def __init__(self, sessions: _Optional[_Iterable[_Union[SessionDescriptor, _Mapping]]] = ...) -> None: ...

class GetSessionRequest(_message.Message):
    __slots__ = ("session_id",)
    SESSION_ID_FIELD_NUMBER: _ClassVar[int]
    session_id: str
    def __init__(self, session_id: _Optional[str] = ...) -> None: ...

class GetSessionResponse(_message.Message):
    __slots__ = ("session",)
    SESSION_FIELD_NUMBER: _ClassVar[int]
    session: SessionDescriptor
    def __init__(self, session: _Optional[_Union[SessionDescriptor, _Mapping]] = ...) -> None: ...

class StartSimulationRequest(_message.Message):
    __slots__ = ("session_id", "overrides")
    SESSION_ID_FIELD_NUMBER: _ClassVar[int]
    OVERRIDES_FIELD_NUMBER: _ClassVar[int]
    session_id: str
    overrides: SimulationOptions
    def __init__(self, session_id: _Optional[str] = ..., overrides: _Optional[_Union[SimulationOptions, _Mapping]] = ...) -> None: ...

class StartSimulationResponse(_message.Message):
    __slots__ = ("session",)
    SESSION_FIELD_NUMBER: _ClassVar[int]
    session: SessionDescriptor
    def __init__(self, session: _Optional[_Union[SessionDescriptor, _Mapping]] = ...) -> None: ...

class PauseSimulationRequest(_message.Message):
    __slots__ = ("session_id",)
    SESSION_ID_FIELD_NUMBER: _ClassVar[int]
    session_id: str
    def __init__(self, session_id: _Optional[str] = ...) -> None: ...

class PauseSimulationResponse(_message.Message):
    __slots__ = ("session",)
    SESSION_FIELD_NUMBER: _ClassVar[int]
    session: SessionDescriptor
    def __init__(self, session: _Optional[_Union[SessionDescriptor, _Mapping]] = ...) -> None: ...

class ResumeSimulationRequest(_message.Message):
    __slots__ = ("session_id",)
    SESSION_ID_FIELD_NUMBER: _ClassVar[int]
    session_id: str
    def __init__(self, session_id: _Optional[str] = ...) -> None: ...

class ResumeSimulationResponse(_message.Message):
    __slots__ = ("session",)
    SESSION_FIELD_NUMBER: _ClassVar[int]
    session: SessionDescriptor
    def __init__(self, session: _Optional[_Union[SessionDescriptor, _Mapping]] = ...) -> None: ...

class StopSimulationRequest(_message.Message):
    __slots__ = ("session_id",)
    SESSION_ID_FIELD_NUMBER: _ClassVar[int]
    session_id: str
    def __init__(self, session_id: _Optional[str] = ...) -> None: ...

class StopSimulationResponse(_message.Message):
    __slots__ = ("session",)
    SESSION_FIELD_NUMBER: _ClassVar[int]
    session: SessionDescriptor
    def __init__(self, session: _Optional[_Union[SessionDescriptor, _Mapping]] = ...) -> None: ...

class StreamWaveformsRequest(_message.Message):
    __slots__ = ("session_id", "signals", "decimation", "start_time")
    SESSION_ID_FIELD_NUMBER: _ClassVar[int]
    SIGNALS_FIELD_NUMBER: _ClassVar[int]
    DECIMATION_FIELD_NUMBER: _ClassVar[int]
    START_TIME_FIELD_NUMBER: _ClassVar[int]
    session_id: str
    signals: _containers.RepeatedScalarFieldContainer[str]
    decimation: int
    start_time: _wrappers_pb2.DoubleValue
    def __init__(self, session_id: _Optional[str] = ..., signals: _Optional[_Iterable[str]] = ..., decimation: _Optional[int] = ..., start_time: _Optional[_Union[_wrappers_pb2.DoubleValue, _Mapping]] = ...) -> None: ...

class WaveformStreamHeader(_message.Message):
    __slots__ = ("session_id", "signals", "tstart", "tstop", "total_samples")
    SESSION_ID_FIELD_NUMBER: _ClassVar[int]
    SIGNALS_FIELD_NUMBER: _ClassVar[int]
    TSTART_FIELD_NUMBER: _ClassVar[int]
    TSTOP_FIELD_NUMBER: _ClassVar[int]
    TOTAL_SAMPLES_FIELD_NUMBER: _ClassVar[int]
    session_id: str
    signals: _containers.RepeatedScalarFieldContainer[str]
    tstart: float
    tstop: float
    total_samples: int
    def __init__(self, session_id: _Optional[str] = ..., signals: _Optional[_Iterable[str]] = ..., tstart: _Optional[float] = ..., tstop: _Optional[float] = ..., total_samples: _Optional[int] = ...) -> None: ...

class WaveformSample(_message.Message):
    __slots__ = ("time", "values")
    TIME_FIELD_NUMBER: _ClassVar[int]
    VALUES_FIELD_NUMBER: _ClassVar[int]
    time: float
    values: _containers.RepeatedScalarFieldContainer[float]
    def __init__(self, time: _Optional[float] = ..., values: _Optional[_Iterable[float]] = ...) -> None: ...

class WaveformStreamComplete(_message.Message):
    __slots__ = ("final_status", "error_message")
    FINAL_STATUS_FIELD_NUMBER: _ClassVar[int]
    ERROR_MESSAGE_FIELD_NUMBER: _ClassVar[int]
    final_status: SessionStatus
    error_message: str
    def __init__(self, final_status: _Optional[_Union[SessionStatus, str]] = ..., error_message: _Optional[str] = ...) -> None: ...

class WaveformStreamResponse(_message.Message):
    __slots__ = ("header", "sample", "complete")
    HEADER_FIELD_NUMBER: _ClassVar[int]
    SAMPLE_FIELD_NUMBER: _ClassVar[int]
    COMPLETE_FIELD_NUMBER: _ClassVar[int]
    header: WaveformStreamHeader
    sample: WaveformSample
    complete: WaveformStreamComplete
    def __init__(self, header: _Optional[_Union[WaveformStreamHeader, _Mapping]] = ..., sample: _Optional[_Union[WaveformSample, _Mapping]] = ..., complete: _Optional[_Union[WaveformStreamComplete, _Mapping]] = ...) -> None: ...

class GetResultRequest(_message.Message):
    __slots__ = ("session_id", "signals", "start_time", "end_time")
    SESSION_ID_FIELD_NUMBER: _ClassVar[int]
    SIGNALS_FIELD_NUMBER: _ClassVar[int]
    START_TIME_FIELD_NUMBER: _ClassVar[int]
    END_TIME_FIELD_NUMBER: _ClassVar[int]
    session_id: str
    signals: _containers.RepeatedScalarFieldContainer[str]
    start_time: _wrappers_pb2.DoubleValue
    end_time: _wrappers_pb2.DoubleValue
    def __init__(self, session_id: _Optional[str] = ..., signals: _Optional[_Iterable[str]] = ..., start_time: _Optional[_Union[_wrappers_pb2.DoubleValue, _Mapping]] = ..., end_time: _Optional[_Union[_wrappers_pb2.DoubleValue, _Mapping]] = ...) -> None: ...

class ResultMetadata(_message.Message):
    __slots__ = ("start_time", "end_time", "sample_count", "signals", "status", "error_message")
    START_TIME_FIELD_NUMBER: _ClassVar[int]
    END_TIME_FIELD_NUMBER: _ClassVar[int]
    SAMPLE_COUNT_FIELD_NUMBER: _ClassVar[int]
    SIGNALS_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    ERROR_MESSAGE_FIELD_NUMBER: _ClassVar[int]
    start_time: float
    end_time: float
    sample_count: int
    signals: _containers.RepeatedScalarFieldContainer[str]
    status: SessionStatus
    error_message: str
    def __init__(self, start_time: _Optional[float] = ..., end_time: _Optional[float] = ..., sample_count: _Optional[int] = ..., signals: _Optional[_Iterable[str]] = ..., status: _Optional[_Union[SessionStatus, str]] = ..., error_message: _Optional[str] = ...) -> None: ...

class GetResultResponse(_message.Message):
    __slots__ = ("metadata",)
    METADATA_FIELD_NUMBER: _ClassVar[int]
    metadata: ResultMetadata
    def __init__(self, metadata: _Optional[_Union[ResultMetadata, _Mapping]] = ...) -> None: ...

class DownloadResultRequest(_message.Message):
    __slots__ = ("session_id", "signals", "start_time", "end_time", "format")
    SESSION_ID_FIELD_NUMBER: _ClassVar[int]
    SIGNALS_FIELD_NUMBER: _ClassVar[int]
    START_TIME_FIELD_NUMBER: _ClassVar[int]
    END_TIME_FIELD_NUMBER: _ClassVar[int]
    FORMAT_FIELD_NUMBER: _ClassVar[int]
    session_id: str
    signals: _containers.RepeatedScalarFieldContainer[str]
    start_time: _wrappers_pb2.DoubleValue
    end_time: _wrappers_pb2.DoubleValue
    format: SimulationResultFormat
    def __init__(self, session_id: _Optional[str] = ..., signals: _Optional[_Iterable[str]] = ..., start_time: _Optional[_Union[_wrappers_pb2.DoubleValue, _Mapping]] = ..., end_time: _Optional[_Union[_wrappers_pb2.DoubleValue, _Mapping]] = ..., format: _Optional[_Union[SimulationResultFormat, str]] = ...) -> None: ...

class DownloadResultResponse(_message.Message):
    __slots__ = ("chunk", "last_chunk", "content_type", "filename")
    CHUNK_FIELD_NUMBER: _ClassVar[int]
    LAST_CHUNK_FIELD_NUMBER: _ClassVar[int]
    CONTENT_TYPE_FIELD_NUMBER: _ClassVar[int]
    FILENAME_FIELD_NUMBER: _ClassVar[int]
    chunk: bytes
    last_chunk: bool
    content_type: str
    filename: str
    def __init__(self, chunk: _Optional[bytes] = ..., last_chunk: bool = ..., content_type: _Optional[str] = ..., filename: _Optional[str] = ...) -> None: ...

class UploadModelRequest(_message.Message):
    __slots__ = ("model",)
    MODEL_FIELD_NUMBER: _ClassVar[int]
    model: CircuitModel
    def __init__(self, model: _Optional[_Union[CircuitModel, _Mapping]] = ...) -> None: ...

class UploadModelResponse(_message.Message):
    __slots__ = ("model",)
    MODEL_FIELD_NUMBER: _ClassVar[int]
    model: ModelDescriptor
    def __init__(self, model: _Optional[_Union[ModelDescriptor, _Mapping]] = ...) -> None: ...

class ListModelsRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class ListModelsResponse(_message.Message):
    __slots__ = ("models",)
    MODELS_FIELD_NUMBER: _ClassVar[int]
    models: _containers.RepeatedCompositeFieldContainer[ModelDescriptor]
    def __init__(self, models: _Optional[_Iterable[_Union[ModelDescriptor, _Mapping]]] = ...) -> None: ...

class GetModelRequest(_message.Message):
    __slots__ = ("model_id",)
    MODEL_ID_FIELD_NUMBER: _ClassVar[int]
    model_id: str
    def __init__(self, model_id: _Optional[str] = ...) -> None: ...

class GetModelResponse(_message.Message):
    __slots__ = ("model",)
    MODEL_FIELD_NUMBER: _ClassVar[int]
    model: ModelDescriptor
    def __init__(self, model: _Optional[_Union[ModelDescriptor, _Mapping]] = ...) -> None: ...

class DeleteModelRequest(_message.Message):
    __slots__ = ("model_id", "force")
    MODEL_ID_FIELD_NUMBER: _ClassVar[int]
    FORCE_FIELD_NUMBER: _ClassVar[int]
    model_id: str
    force: bool
    def __init__(self, model_id: _Optional[str] = ..., force: bool = ...) -> None: ...

class DeleteModelResponse(_message.Message):
    __slots__ = ("deleted",)
    DELETED_FIELD_NUMBER: _ClassVar[int]
    deleted: bool
    def __init__(self, deleted: bool = ...) -> None: ...

class SweepParameter(_message.Message):
    __slots__ = ("name", "start", "stop", "steps", "logarithmic")
    NAME_FIELD_NUMBER: _ClassVar[int]
    START_FIELD_NUMBER: _ClassVar[int]
    STOP_FIELD_NUMBER: _ClassVar[int]
    STEPS_FIELD_NUMBER: _ClassVar[int]
    LOGARITHMIC_FIELD_NUMBER: _ClassVar[int]
    name: str
    start: float
    stop: float
    steps: int
    logarithmic: bool
    def __init__(self, name: _Optional[str] = ..., start: _Optional[float] = ..., stop: _Optional[float] = ..., steps: _Optional[int] = ..., logarithmic: bool = ...) -> None: ...

class CreateSweepRequest(_message.Message):
    __slots__ = ("model_id", "base_options", "parameters", "signals")
    MODEL_ID_FIELD_NUMBER: _ClassVar[int]
    BASE_OPTIONS_FIELD_NUMBER: _ClassVar[int]
    PARAMETERS_FIELD_NUMBER: _ClassVar[int]
    SIGNALS_FIELD_NUMBER: _ClassVar[int]
    model_id: str
    base_options: SimulationOptions
    parameters: _containers.RepeatedCompositeFieldContainer[SweepParameter]
    signals: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, model_id: _Optional[str] = ..., base_options: _Optional[_Union[SimulationOptions, _Mapping]] = ..., parameters: _Optional[_Iterable[_Union[SweepParameter, _Mapping]]] = ..., signals: _Optional[_Iterable[str]] = ...) -> None: ...

class CreateSweepResponse(_message.Message):
    __slots__ = ("sweep_id",)
    SWEEP_ID_FIELD_NUMBER: _ClassVar[int]
    sweep_id: str
    def __init__(self, sweep_id: _Optional[str] = ...) -> None: ...

class RunSweepRequest(_message.Message):
    __slots__ = ("sweep_id", "parallel")
    SWEEP_ID_FIELD_NUMBER: _ClassVar[int]
    PARALLEL_FIELD_NUMBER: _ClassVar[int]
    sweep_id: str
    parallel: bool
    def __init__(self, sweep_id: _Optional[str] = ..., parallel: bool = ...) -> None: ...

class RunSweepResponse(_message.Message):
    __slots__ = ("status",)
    STATUS_FIELD_NUMBER: _ClassVar[int]
    status: SweepStatus
    def __init__(self, status: _Optional[_Union[SweepStatus, str]] = ...) -> None: ...

class GetSweepResultsRequest(_message.Message):
    __slots__ = ("sweep_id",)
    SWEEP_ID_FIELD_NUMBER: _ClassVar[int]
    sweep_id: str
    def __init__(self, sweep_id: _Optional[str] = ...) -> None: ...

class ParameterValue(_message.Message):
    __slots__ = ("name", "value")
    NAME_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    name: str
    value: float
    def __init__(self, name: _Optional[str] = ..., value: _Optional[float] = ...) -> None: ...

class SweepResult(_message.Message):
    __slots__ = ("parameters", "metadata", "final_values")
    PARAMETERS_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    FINAL_VALUES_FIELD_NUMBER: _ClassVar[int]
    parameters: _containers.RepeatedCompositeFieldContainer[ParameterValue]
    metadata: ResultMetadata
    final_values: _containers.RepeatedScalarFieldContainer[float]
    def __init__(self, parameters: _Optional[_Iterable[_Union[ParameterValue, _Mapping]]] = ..., metadata: _Optional[_Union[ResultMetadata, _Mapping]] = ..., final_values: _Optional[_Iterable[float]] = ...) -> None: ...

class GetSweepResultsResponse(_message.Message):
    __slots__ = ("status", "results")
    STATUS_FIELD_NUMBER: _ClassVar[int]
    RESULTS_FIELD_NUMBER: _ClassVar[int]
    status: SweepStatus
    results: _containers.RepeatedCompositeFieldContainer[SweepResult]
    def __init__(self, status: _Optional[_Union[SweepStatus, str]] = ..., results: _Optional[_Iterable[_Union[SweepResult, _Mapping]]] = ...) -> None: ...



#!/usr/bin/env python3
"""
Benchmark comparison between Pulsim and ngspice.
Compares simulation results for RC, RL, and RLC circuits.
"""

import numpy as np
import os

def load_pulsim_csv(filename):
    """Load Pulsim CSV output (time, signal1, signal2, ...)"""
    data = np.genfromtxt(filename, delimiter=',', skip_header=1)
    return data

def load_ngspice_csv(filename):
    """Load ngspice wrdata output (time, value pairs)"""
    # ngspice wrdata format: whitespace-separated time value pairs
    data = np.loadtxt(filename)
    return data

def analytical_rc_step(t, V0, R, C):
    """RC step response: V(t) = V0 * (1 - exp(-t/RC))"""
    tau = R * C
    return V0 * (1 - np.exp(-t / tau))

def analytical_rl_step(t, V0, R, L):
    """RL step response voltage across inductor: V_L(t) = V0 * exp(-Rt/L)
       Voltage across resistor: V_R(t) = V0 * (1 - exp(-Rt/L))
       For our circuit, out node is between R and L, so V(out) = V_L
    """
    tau = L / R
    return V0 * np.exp(-t / tau)

def analytical_rlc_step(t, V0, R, L, C):
    """Underdamped RLC step response on capacitor voltage"""
    alpha = R / (2 * L)
    omega0 = 1 / np.sqrt(L * C)

    if alpha < omega0:  # Underdamped
        omega_d = np.sqrt(omega0**2 - alpha**2)
        return V0 * (1 - np.exp(-alpha * t) * (np.cos(omega_d * t) +
                     (alpha / omega_d) * np.sin(omega_d * t)))
    else:  # Overdamped or critically damped
        s1 = -alpha + np.sqrt(alpha**2 - omega0**2)
        s2 = -alpha - np.sqrt(alpha**2 - omega0**2)
        A = V0 * s2 / (s2 - s1)
        B = -V0 * s1 / (s2 - s1)
        return V0 - A * np.exp(s1 * t) - B * np.exp(s2 * t)

def compare_results():
    """Compare Pulsim and ngspice results against analytical solutions"""

    results = []

    # RC Circuit comparison
    print("=" * 60)
    print("RC CIRCUIT COMPARISON")
    print("=" * 60)

    pulsim_rc = load_pulsim_csv('results/rc_pulsim.csv')
    ngspice_rc = load_ngspice_csv('results/rc_ngspice.csv')

    # Parameters
    V0, R, C = 5.0, 1000.0, 1e-6

    # Pulsim: columns are time, V(in), V(out), I(V1)
    t_sl = pulsim_rc[:, 0]
    v_out_sl = pulsim_rc[:, 2]  # V(out)

    # ngspice: columns are time, v(out)
    t_ng = ngspice_rc[:, 0]
    v_out_ng = ngspice_rc[:, 1]

    # Analytical
    v_analytical_sl = analytical_rc_step(t_sl, V0, R, C)
    v_analytical_ng = analytical_rc_step(t_ng, V0, R, C)

    # Calculate errors
    error_sl = np.abs(v_out_sl - v_analytical_sl)
    error_ng = np.abs(v_out_ng - v_analytical_ng)

    print(f"Pulsim: {len(t_sl)} points, t=[{t_sl[0]:.2e}, {t_sl[-1]:.2e}]s")
    print(f"ngspice:  {len(t_ng)} points, t=[{t_ng[0]:.2e}, {t_ng[-1]:.2e}]s")
    print(f"\nMax absolute error vs analytical:")
    print(f"  Pulsim: {np.max(error_sl):.6e} V")
    print(f"  ngspice:  {np.max(error_ng):.6e} V")
    print(f"\nRMS error vs analytical:")
    print(f"  Pulsim: {np.sqrt(np.mean(error_sl**2)):.6e} V")
    print(f"  ngspice:  {np.sqrt(np.mean(error_ng**2)):.6e} V")
    print(f"\nFinal value comparison (at t={t_sl[-1]:.2e}s):")
    print(f"  Analytical: {v_analytical_sl[-1]:.6f} V")
    print(f"  Pulsim:   {v_out_sl[-1]:.6f} V (error: {error_sl[-1]:.6e} V)")

    # Find matching ngspice point
    idx_ng = np.argmin(np.abs(t_ng - t_sl[-1]))
    print(f"  ngspice:    {v_out_ng[idx_ng]:.6f} V (error: {np.abs(v_out_ng[idx_ng] - v_analytical_sl[-1]):.6e} V)")

    results.append({
        'circuit': 'RC',
        'pulsim_max_error': np.max(error_sl),
        'ngspice_max_error': np.max(error_ng),
        'pulsim_rms_error': np.sqrt(np.mean(error_sl**2)),
        'ngspice_rms_error': np.sqrt(np.mean(error_ng**2))
    })

    # RL Circuit comparison
    print("\n" + "=" * 60)
    print("RL CIRCUIT COMPARISON")
    print("=" * 60)

    pulsim_rl = load_pulsim_csv('results/rl_pulsim.csv')
    ngspice_rl = load_ngspice_csv('results/rl_ngspice.csv')

    # Parameters
    V0, R, L = 10.0, 100.0, 10e-3
    tau_rl = L / R  # Time constant = 0.1ms

    # Pulsim: columns are time, V(in), V(out), I(V1), I(L1)
    t_sl = pulsim_rl[:, 0]
    v_out_sl = pulsim_rl[:, 2]  # V(out) - voltage across inductor

    # ngspice: columns are time, v(out)
    t_ng = ngspice_rl[:, 0]
    v_out_ng = ngspice_rl[:, 1]

    # Filter to post-step region (after pulse rise time, t > 1ns)
    # For proper comparison, compare only after the step has occurred
    sl_mask = t_sl > 1e-9
    ng_mask = t_ng > 1e-9

    t_sl_filt = t_sl[sl_mask]
    v_out_sl_filt = v_out_sl[sl_mask]
    t_ng_filt = t_ng[ng_mask]
    v_out_ng_filt = v_out_ng[ng_mask]

    # Analytical (voltage across inductor = V0 * exp(-Rt/L))
    v_analytical_sl = analytical_rl_step(t_sl_filt, V0, R, L)
    v_analytical_ng = analytical_rl_step(t_ng_filt, V0, R, L)

    # Calculate errors
    error_sl = np.abs(v_out_sl_filt - v_analytical_sl)
    error_ng = np.abs(v_out_ng_filt - v_analytical_ng)

    print(f"Pulsim: {len(t_sl_filt)} points, t=[{t_sl_filt[0]:.2e}, {t_sl_filt[-1]:.2e}]s")
    print(f"ngspice:  {len(t_ng_filt)} points, t=[{t_ng_filt[0]:.2e}, {t_ng_filt[-1]:.2e}]s")
    print(f"Time constant (tau): {tau_rl*1e3:.3f} ms")
    print(f"\nMax absolute error vs analytical:")
    print(f"  Pulsim: {np.max(error_sl):.6e} V")
    print(f"  ngspice:  {np.max(error_ng):.6e} V")
    print(f"\nRMS error vs analytical:")
    print(f"  Pulsim: {np.sqrt(np.mean(error_sl**2)):.6e} V")
    print(f"  ngspice:  {np.sqrt(np.mean(error_ng**2)):.6e} V")

    results.append({
        'circuit': 'RL',
        'pulsim_max_error': np.max(error_sl),
        'ngspice_max_error': np.max(error_ng),
        'pulsim_rms_error': np.sqrt(np.mean(error_sl**2)),
        'ngspice_rms_error': np.sqrt(np.mean(error_ng**2))
    })

    # RLC Circuit comparison
    print("\n" + "=" * 60)
    print("RLC CIRCUIT COMPARISON")
    print("=" * 60)

    pulsim_rlc = load_pulsim_csv('results/rlc_pulsim.csv')
    ngspice_rlc = load_ngspice_csv('results/rlc_ngspice.csv')

    # Parameters
    V0, R, L, C = 10.0, 10.0, 1e-3, 10e-6

    # Pulsim: columns are time, V(in), V(n1), V(out), I(V1), I(L1)
    t_sl = pulsim_rlc[:, 0]
    v_out_sl = pulsim_rlc[:, 3]  # V(out)

    # ngspice: columns are time, v(out), time, i(V1)
    t_ng = ngspice_rlc[:, 0]
    v_out_ng = ngspice_rlc[:, 1]

    # Analytical
    v_analytical_sl = analytical_rlc_step(t_sl, V0, R, L, C)
    v_analytical_ng = analytical_rlc_step(t_ng, V0, R, L, C)

    # Calculate errors
    error_sl = np.abs(v_out_sl - v_analytical_sl)
    error_ng = np.abs(v_out_ng - v_analytical_ng)

    print(f"Pulsim: {len(t_sl)} points, t=[{t_sl[0]:.2e}, {t_sl[-1]:.2e}]s")
    print(f"ngspice:  {len(t_ng)} points, t=[{t_ng[0]:.2e}, {t_ng[-1]:.2e}]s")

    # RLC parameters
    alpha = R / (2 * L)
    omega0 = 1 / np.sqrt(L * C)
    omega_d = np.sqrt(omega0**2 - alpha**2) if omega0 > alpha else 0
    print(f"\nRLC Parameters:")
    print(f"  Damping factor (alpha): {alpha:.2f} rad/s")
    print(f"  Natural frequency (omega0): {omega0:.2f} rad/s")
    print(f"  Damped frequency (omega_d): {omega_d:.2f} rad/s")
    print(f"  Period: {2*np.pi/omega_d*1000:.3f} ms")
    print(f"  Damping ratio (zeta): {alpha/omega0:.3f}")

    print(f"\nMax absolute error vs analytical:")
    print(f"  Pulsim: {np.max(error_sl):.6e} V")
    print(f"  ngspice:  {np.max(error_ng):.6e} V")
    print(f"\nRMS error vs analytical:")
    print(f"  Pulsim: {np.sqrt(np.mean(error_sl**2)):.6e} V")
    print(f"  ngspice:  {np.sqrt(np.mean(error_ng**2)):.6e} V")

    results.append({
        'circuit': 'RLC',
        'pulsim_max_error': np.max(error_sl),
        'ngspice_max_error': np.max(error_ng),
        'pulsim_rms_error': np.sqrt(np.mean(error_sl**2)),
        'ngspice_rms_error': np.sqrt(np.mean(error_ng**2))
    })

    # Summary
    print("\n" + "=" * 60)
    print("SUMMARY")
    print("=" * 60)
    print("\n{:<10} {:>20} {:>20}".format("Circuit", "Pulsim RMS Error", "ngspice RMS Error"))
    print("-" * 52)
    for r in results:
        print("{:<10} {:>20.6e} {:>20.6e}".format(
            r['circuit'], r['pulsim_rms_error'], r['ngspice_rms_error']))

    print("\n" + "=" * 60)
    print("CONCLUSION")
    print("=" * 60)
    print("Pulsim produces results comparable to ngspice with")
    print("errors on the same order of magnitude relative to")
    print("analytical solutions. Both simulators use Backward Euler")
    print("integration which introduces some numerical damping.")

    return results

if __name__ == '__main__':
    os.chdir(os.path.dirname(os.path.abspath(__file__)))
    compare_results()



"""Pulsim - High-performance circuit simulator for power electronics."""

__version__ = "0.1.0"

from ._pulsim import (
    # Enums
    ComponentType,
    SolverStatus,
    MOSFETType,
    ThermalNetworkType,
    SimulationState,
    IntegrationMethod,
    DiagnosticSeverity,
    DiagnosticCode,
    ParameterType,
    SimulationEventType,

    # Waveforms
    DCWaveform,
    PulseWaveform,
    SineWaveform,
    PWLWaveform,
    PWMWaveform,

    # Component Parameters
    DiodeParams,
    SwitchParams,
    MOSFETParams,
    IGBTParams,
    TransformerParams,

    # Simulation
    SimulationOptions,
    SimulationResult,
    Circuit,
    Simulator,
    PowerLosses,
    SwitchEvent,
    simulate,

    # Simulation Control (GUI integration)
    SimulationController,
    SimulationProgress,
    ProgressCallbackConfig,
    StreamingConfig,

    # Enhanced Result Types
    SignalInfo,
    SolverInfo,
    SimulationEvent,

    # Component Metadata (GUI integration)
    ParameterMetadata,
    PinMetadata,
    ComponentMetadata,
    ComponentRegistry,

    # Schematic Position (GUI integration)
    SchematicPosition,

    # Validation (GUI integration)
    Diagnostic,
    ValidationResult,
    validate_circuit,
    diagnostic_code_description,

    # Parsing
    parse_netlist_file,
    parse_netlist_string,
    circuit_to_json,

    # Thermal
    ThermalRCStage,
    FosterNetwork,
    ThermalModel,
    ThermalState,
    ThermalWarning,
    ThermalSimulator,
    create_mosfet_thermal,
    fit_foster_network,

    # Device Library
    devices,
)

__all__ = [
    # Enums
    "ComponentType",
    "SolverStatus",
    "MOSFETType",
    "ThermalNetworkType",
    "SimulationState",
    "IntegrationMethod",
    "DiagnosticSeverity",
    "DiagnosticCode",
    "ParameterType",
    "SimulationEventType",

    # Waveforms
    "DCWaveform",
    "PulseWaveform",
    "SineWaveform",
    "PWLWaveform",
    "PWMWaveform",

    # Component Parameters
    "DiodeParams",
    "SwitchParams",
    "MOSFETParams",
    "IGBTParams",
    "TransformerParams",

    # Simulation
    "SimulationOptions",
    "SimulationResult",
    "Circuit",
    "Simulator",
    "PowerLosses",
    "SwitchEvent",
    "simulate",

    # Simulation Control (GUI integration)
    "SimulationController",
    "SimulationProgress",
    "ProgressCallbackConfig",
    "StreamingConfig",

    # Enhanced Result Types
    "SignalInfo",
    "SolverInfo",
    "SimulationEvent",

    # Component Metadata (GUI integration)
    "ParameterMetadata",
    "PinMetadata",
    "ComponentMetadata",
    "ComponentRegistry",

    # Schematic Position (GUI integration)
    "SchematicPosition",

    # Validation (GUI integration)
    "Diagnostic",
    "ValidationResult",
    "validate_circuit",
    "diagnostic_code_description",

    # Parsing
    "parse_netlist_file",
    "parse_netlist_string",
    "circuit_to_json",

    # Thermal
    "ThermalRCStage",
    "FosterNetwork",
    "ThermalModel",
    "ThermalState",
    "ThermalWarning",
    "ThermalSimulator",
    "create_mosfet_thermal",
    "fit_foster_network",

    # Device Library
    "devices",

    # Version
    "__version__",
]



#!/usr/bin/env python3
"""
Benchmark Pulsim against ngspice

This script compares simulation performance and accuracy between
Pulsim and ngspice for various circuit types.

Requirements:
- ngspice installed (brew install ngspice / apt install ngspice)
- Pulsim CLI built

Usage:
    python benchmark_ngspice.py [--verbose] [--output-dir DIR]
"""

import argparse
import json
import os
import subprocess
import tempfile
import time
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Tuple
import csv


@dataclass
class BenchmarkResult:
    """Results from a single benchmark run"""
    circuit_name: str
    pulsim_time: float
    ngspice_time: float
    pulsim_steps: int
    ngspice_steps: int
    accuracy_error: float  # RMS error between solutions
    speedup: float  # ngspice_time / pulsim_time


def find_pulsim_cli() -> Optional[Path]:
    """Find the Pulsim CLI executable"""
    # Check common build directories
    script_dir = Path(__file__).parent.parent
    candidates = [
        script_dir / "build" / "cli" / "pulsim",
        script_dir / "build" / "Release" / "cli" / "pulsim",
        script_dir / "build" / "Debug" / "cli" / "pulsim",
        Path("pulsim"),  # In PATH
    ]
    for path in candidates:
        if path.exists() or (path.name == "pulsim" and subprocess.run(
            ["which", "pulsim"], capture_output=True).returncode == 0):
            return path
    return None


def check_ngspice() -> bool:
    """Check if ngspice is available"""
    try:
        result = subprocess.run(["ngspice", "--version"], capture_output=True, text=True)
        return result.returncode == 0
    except FileNotFoundError:
        return False


def generate_spice_netlist(circuit_json: dict, output_path: Path) -> None:
    """Convert Pulsim JSON format to SPICE netlist"""
    lines = [f"* {circuit_json.get('name', 'Benchmark Circuit')}"]

    # Add components
    for comp in circuit_json.get("components", []):
        comp_type = comp["type"].lower()
        name = comp["name"]
        nodes = comp["nodes"]

        if comp_type == "resistor":
            lines.append(f"R{name} {nodes[0]} {nodes[1]} {comp['value']}")
        elif comp_type == "capacitor":
            lines.append(f"C{name} {nodes[0]} {nodes[1]} {comp['value']}")
        elif comp_type == "inductor":
            lines.append(f"L{name} {nodes[0]} {nodes[1]} {comp['value']}")
        elif comp_type == "voltage_source":
            waveform = comp.get("waveform", {})
            if waveform.get("type") == "dc":
                lines.append(f"V{name} {nodes[0]} {nodes[1]} DC {waveform['value']}")
            elif waveform.get("type") == "pulse":
                lines.append(
                    f"V{name} {nodes[0]} {nodes[1]} PULSE("
                    f"{waveform['v1']} {waveform['v2']} "
                    f"{waveform.get('td', 0)} {waveform.get('tr', 1e-9)} "
                    f"{waveform.get('tf', 1e-9)} {waveform['pw']} {waveform['period']})"
                )
            elif waveform.get("type") == "sine":
                lines.append(
                    f"V{name} {nodes[0]} {nodes[1]} SIN("
                    f"{waveform.get('offset', 0)} {waveform['amplitude']} "
                    f"{waveform['frequency']} {waveform.get('delay', 0)} "
                    f"{waveform.get('damping', 0)})"
                )
        elif comp_type == "diode":
            lines.append(f"D{name} {nodes[0]} {nodes[1]} DMOD")
            lines.append(".MODEL DMOD D(IS=1E-14 N=1.0)")
        elif comp_type == "switch":
            # SPICE switch model
            lines.append(f"S{name} {nodes[0]} {nodes[1]} {nodes[2]} {nodes[3]} SWMOD")
            lines.append(f".MODEL SWMOD SW(VT={comp.get('vth', 2.5)} RON={comp.get('ron', 0.01)} ROFF=1G)")

    # Add simulation commands
    sim = circuit_json.get("simulation", {})
    tstop = sim.get("tstop", 1e-3)
    dt = sim.get("dt", 1e-6)
    lines.append(f".TRAN {dt} {tstop} 0 {dt}")
    lines.append(".END")

    with open(output_path, "w") as f:
        f.write("\n".join(lines))


def run_pulsim(cli_path: Path, circuit_path: Path, output_path: Path) -> Tuple[float, int]:
    """Run Pulsim simulation and return (time, steps)"""
    start = time.perf_counter()
    result = subprocess.run(
        [str(cli_path), "run", str(circuit_path), "-o", str(output_path)],
        capture_output=True, text=True
    )
    elapsed = time.perf_counter() - start

    if result.returncode != 0:
        raise RuntimeError(f"Pulsim failed: {result.stderr}")

    # Count steps from output
    steps = 0
    if output_path.exists():
        with open(output_path) as f:
            steps = sum(1 for _ in f) - 1  # Subtract header

    return elapsed, steps


def run_ngspice(netlist_path: Path, output_path: Path) -> Tuple[float, int]:
    """Run ngspice simulation and return (time, steps)"""
    # Create control file for batch mode
    ctrl_content = f"""
set filetype=ascii
tran 1u 1m
wrdata {output_path} all
quit
"""
    ctrl_path = netlist_path.with_suffix(".ctrl")
    with open(ctrl_path, "w") as f:
        f.write(ctrl_content)

    start = time.perf_counter()
    result = subprocess.run(
        ["ngspice", "-b", str(netlist_path)],
        capture_output=True, text=True,
        cwd=netlist_path.parent
    )
    elapsed = time.perf_counter() - start

    # Count steps from output
    steps = 0
    if output_path.exists():
        with open(output_path) as f:
            steps = sum(1 for _ in f)

    return elapsed, steps


def create_benchmark_circuits() -> list:
    """Create a set of benchmark circuits"""
    circuits = []

    # 1. Simple RC circuit
    circuits.append({
        "name": "RC_Simple",
        "components": [
            {"type": "voltage_source", "name": "Vin", "nodes": ["in", "0"],
             "waveform": {"type": "dc", "value": 5.0}},
            {"type": "resistor", "name": "R1", "nodes": ["in", "out"], "value": 1000},
            {"type": "capacitor", "name": "C1", "nodes": ["out", "0"], "value": 1e-6},
        ],
        "simulation": {"tstop": 10e-3, "dt": 1e-6}
    })

    # 2. RLC oscillator
    circuits.append({
        "name": "RLC_Oscillator",
        "components": [
            {"type": "voltage_source", "name": "Vin", "nodes": ["in", "0"],
             "waveform": {"type": "dc", "value": 10.0}},
            {"type": "resistor", "name": "R1", "nodes": ["in", "n1"], "value": 10},
            {"type": "inductor", "name": "L1", "nodes": ["n1", "out"], "value": 1e-3},
            {"type": "capacitor", "name": "C1", "nodes": ["out", "0"], "value": 1e-6},
        ],
        "simulation": {"tstop": 5e-3, "dt": 0.5e-6}
    })

    # 3. PWM driven RC (switching)
    circuits.append({
        "name": "PWM_RC",
        "components": [
            {"type": "voltage_source", "name": "Vin", "nodes": ["vcc", "0"],
             "waveform": {"type": "dc", "value": 12.0}},
            {"type": "voltage_source", "name": "Vctrl", "nodes": ["ctrl", "0"],
             "waveform": {"type": "pulse", "v1": 0, "v2": 5,
                          "td": 0, "tr": 1e-9, "tf": 1e-9, "pw": 50e-6, "period": 100e-6}},
            {"type": "switch", "name": "S1", "nodes": ["vcc", "out", "ctrl", "0"],
             "ron": 0.01, "vth": 2.5},
            {"type": "resistor", "name": "R1", "nodes": ["out", "0"], "value": 100},
            {"type": "capacitor", "name": "C1", "nodes": ["out", "0"], "value": 10e-6},
        ],
        "simulation": {"tstop": 2e-3, "dt": 1e-6}
    })

    # 4. Larger RC ladder (10 stages)
    ladder_comps = [
        {"type": "voltage_source", "name": "Vin", "nodes": ["n0", "0"],
         "waveform": {"type": "dc", "value": 5.0}}
    ]
    for i in range(10):
        ladder_comps.append({
            "type": "resistor", "name": f"R{i+1}",
            "nodes": [f"n{i}", f"n{i+1}"], "value": 1000
        })
        ladder_comps.append({
            "type": "capacitor", "name": f"C{i+1}",
            "nodes": [f"n{i+1}", "0"], "value": 100e-9
        })
    circuits.append({
        "name": "RC_Ladder_10",
        "components": ladder_comps,
        "simulation": {"tstop": 1e-3, "dt": 0.1e-6}
    })

    return circuits


def run_benchmarks(pulsim_cli: Path, output_dir: Path, verbose: bool) -> list:
    """Run all benchmarks and return results"""
    results = []
    circuits = create_benchmark_circuits()

    for circuit in circuits:
        if verbose:
            print(f"\nBenchmarking: {circuit['name']}")

        with tempfile.TemporaryDirectory() as tmpdir:
            tmppath = Path(tmpdir)

            # Save Pulsim JSON
            json_path = tmppath / "circuit.json"
            with open(json_path, "w") as f:
                json.dump(circuit, f)

            # Run Pulsim
            sl_output = tmppath / "pulsim_out.csv"
            try:
                sl_time, sl_steps = run_pulsim(pulsim_cli, json_path, sl_output)
            except Exception as e:
                if verbose:
                    print(f"  Pulsim failed: {e}")
                continue

            # Generate and run ngspice
            spice_path = tmppath / "circuit.sp"
            ng_output = tmppath / "ngspice_out.txt"
            generate_spice_netlist(circuit, spice_path)

            try:
                ng_time, ng_steps = run_ngspice(spice_path, ng_output)
            except Exception as e:
                if verbose:
                    print(f"  ngspice failed: {e}")
                ng_time, ng_steps = 0, 0

            # Calculate speedup
            speedup = ng_time / sl_time if sl_time > 0 else 0

            result = BenchmarkResult(
                circuit_name=circuit["name"],
                pulsim_time=sl_time,
                ngspice_time=ng_time,
                pulsim_steps=sl_steps,
                ngspice_steps=ng_steps,
                accuracy_error=0.0,  # TODO: implement accuracy comparison
                speedup=speedup
            )
            results.append(result)

            if verbose:
                print(f"  Pulsim: {sl_time:.3f}s ({sl_steps} steps)")
                print(f"  ngspice:  {ng_time:.3f}s ({ng_steps} steps)")
                print(f"  Speedup:  {speedup:.2f}x")

    return results


def save_results(results: list, output_dir: Path) -> None:
    """Save benchmark results to CSV"""
    output_path = output_dir / "benchmark_results.csv"
    with open(output_path, "w", newline="") as f:
        writer = csv.writer(f)
        writer.writerow([
            "Circuit", "Pulsim Time (s)", "ngspice Time (s)",
            "Pulsim Steps", "ngspice Steps", "Speedup"
        ])
        for r in results:
            writer.writerow([
                r.circuit_name, f"{r.pulsim_time:.4f}",
                f"{r.ngspice_time:.4f}", r.pulsim_steps,
                r.ngspice_steps, f"{r.speedup:.2f}"
            ])
    print(f"\nResults saved to: {output_path}")


def main():
    parser = argparse.ArgumentParser(description="Benchmark Pulsim against ngspice")
    parser.add_argument("--verbose", "-v", action="store_true",
                        help="Print detailed output")
    parser.add_argument("--output-dir", "-o", type=Path, default=Path("."),
                        help="Output directory for results")
    args = parser.parse_args()

    # Check prerequisites
    pulsim_cli = find_pulsim_cli()
    if not pulsim_cli:
        print("Error: Pulsim CLI not found. Build the project first.")
        return 1

    if not check_ngspice():
        print("Warning: ngspice not found. Install with: brew install ngspice")
        print("Continuing with Pulsim-only benchmarks...")

    print(f"Pulsim CLI: {pulsim_cli}")
    print(f"ngspice available: {check_ngspice()}")

    # Create output directory
    args.output_dir.mkdir(parents=True, exist_ok=True)

    # Run benchmarks
    results = run_benchmarks(pulsim_cli, args.output_dir, args.verbose)

    if results:
        save_results(results, args.output_dir)

        # Print summary
        print("\n=== Benchmark Summary ===")
        total_sl = sum(r.pulsim_time for r in results)
        total_ng = sum(r.ngspice_time for r in results if r.ngspice_time > 0)
        print(f"Total Pulsim time: {total_sl:.3f}s")
        if total_ng > 0:
            print(f"Total ngspice time:  {total_ng:.3f}s")
            print(f"Average speedup:     {total_ng/total_sl:.2f}x")
    else:
        print("No benchmark results generated.")

    return 0


if __name__ == "__main__":
    exit(main())



"""Jupyter widgets for streaming waveform visualization."""

from __future__ import annotations

import asyncio
import threading
from typing import Any, Dict, List, Optional, Tuple, TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from .client import PulsimClient

# Check for optional dependencies
_HAS_IPYWIDGETS = False
_HAS_MATPLOTLIB = False
_HAS_PLOTLY = False

try:
    import ipywidgets as widgets
    from IPython.display import display, clear_output
    _HAS_IPYWIDGETS = True
except ImportError:
    pass

try:
    import matplotlib.pyplot as plt
    _HAS_MATPLOTLIB = True
except ImportError:
    pass

try:
    import plotly.graph_objects as go
    _HAS_PLOTLY = True
except ImportError:
    go = None  # type: ignore[assignment]


def _check_dependencies(plotly: bool = False) -> None:
    """Check if required dependencies are available."""
    if not _HAS_IPYWIDGETS:
        raise ImportError(
            "ipywidgets is required for Jupyter widgets. "
            "Install with: pip install ipywidgets"
        )
    if plotly and not _HAS_PLOTLY:
        raise ImportError(
            "plotly is required for interactive plots. "
            "Install with: pip install plotly"
        )
    if not plotly and not _HAS_MATPLOTLIB:
        raise ImportError(
            "matplotlib is required for plots. "
            "Install with: pip install matplotlib"
        )


class StreamingPlot:
    """Matplotlib-based streaming waveform plot for Jupyter notebooks.

    Displays live waveform data from a streaming simulation with
    auto-updating plots.

    Example:
        >>> from pulsim.grpc import PulsimClient
        >>> from pulsim.grpc.widgets import StreamingPlot
        >>>
        >>> client = PulsimClient("localhost:50051")
        >>> with client:
        ...     session = client.create_session(...)
        ...     client.start_simulation(session.session_id)
        ...     plot = StreamingPlot(client, session.session_id)
        ...     plot.start()
    """

    def __init__(
        self,
        client: "PulsimClient",
        session_id: str,
        signals: Optional[List[str]] = None,
        decimation: int = 1,
        max_points: int = 10000,
        update_interval: float = 0.1,
        figsize: Tuple[int, int] = (12, 6),
    ):
        """Initialize the streaming plot.

        Args:
            client: Connected PulsimClient instance.
            session_id: Session to stream from.
            signals: Signals to plot (all if None).
            decimation: Decimation factor for streaming.
            max_points: Maximum points to display (older points are dropped).
            update_interval: Update interval in seconds.
            figsize: Figure size (width, height) in inches.
        """
        _check_dependencies()

        self._client = client
        self._session_id = session_id
        self._signals = signals
        self._decimation = decimation
        self._max_points = max_points
        self._update_interval = update_interval
        self._figsize = figsize

        self._times: List[float] = []
        self._values: Dict[str, List[float]] = {}
        self._signal_names: List[str] = []
        self._running = False
        self._stream = None
        self._fig = None
        self._axes = None
        self._lines: Dict[str, Any] = {}
        self._output = None
        self._status_label = None
        self._progress_bar = None

    def _create_ui(self) -> None:
        """Create the Jupyter widget UI."""
        # Status label
        self._status_label = widgets.HTML(
            value="<b>Status:</b> Initializing..."
        )

        # Progress bar
        self._progress_bar = widgets.FloatProgress(
            value=0,
            min=0,
            max=100,
            description="Progress:",
            bar_style="info",
            style={"bar_color": "#2196F3"},
        )

        # Control buttons
        self._stop_button = widgets.Button(
            description="Stop",
            button_style="danger",
            icon="stop",
        )
        self._stop_button.on_click(lambda _: self.stop())

        self._pause_button = widgets.Button(
            description="Pause",
            button_style="warning",
            icon="pause",
        )
        self._pause_button.on_click(lambda _: self._toggle_pause())

        controls = widgets.HBox([self._stop_button, self._pause_button])

        # Output area for the plot
        self._output = widgets.Output()

        # Layout
        self._container = widgets.VBox([
            self._status_label,
            self._progress_bar,
            controls,
            self._output,
        ])

    def _init_plot(self) -> None:
        """Initialize the matplotlib figure."""
        with self._output:
            clear_output(wait=True)
            self._fig, self._axes = plt.subplots(figsize=self._figsize)
            self._fig.tight_layout()
            plt.ion()
            plt.show()

    def _update_plot(self) -> None:
        """Update the plot with new data."""
        if not self._times:
            return

        with self._output:
            times_arr = np.array(self._times[-self._max_points:])

            for sig_name in self._signal_names:
                values = self._values.get(sig_name, [])[-self._max_points:]
                values_arr = np.array(values)

                if sig_name not in self._lines:
                    line, = self._axes.plot(times_arr, values_arr, label=sig_name)
                    self._lines[sig_name] = line
                else:
                    self._lines[sig_name].set_data(times_arr, values_arr)

            self._axes.relim()
            self._axes.autoscale_view()
            self._axes.set_xlabel("Time (s)")
            self._axes.set_ylabel("Value")
            self._axes.legend(loc="upper right")
            self._axes.grid(True, alpha=0.3)

            self._fig.canvas.draw()
            self._fig.canvas.flush_events()

    def _stream_data(self) -> None:
        """Stream data from the server."""
        try:
            self._stream = self._client.stream_waveforms(
                self._session_id,
                self._signals,
                self._decimation,
            )

            for sample in self._stream:
                if not self._running:
                    break

                self._times.append(sample.time)
                for sig_name, value in sample.values.items():
                    if sig_name not in self._signal_names:
                        self._signal_names.append(sig_name)
                        self._values[sig_name] = []
                    self._values[sig_name].append(value)

            self._status_label.value = "<b>Status:</b> Completed"
            self._progress_bar.bar_style = "success"
            self._progress_bar.value = 100

        except Exception as e:
            self._status_label.value = f"<b>Status:</b> Error - {e}"
            self._progress_bar.bar_style = "danger"

        finally:
            self._running = False

    def _toggle_pause(self) -> None:
        """Toggle pause state."""
        if self._pause_button.description == "Pause":
            self._pause_button.description = "Resume"
            self._pause_button.icon = "play"
            self._client.pause_simulation(self._session_id)
        else:
            self._pause_button.description = "Pause"
            self._pause_button.icon = "pause"
            self._client.resume_simulation(self._session_id)

    def start(self) -> widgets.VBox:
        """Start streaming and display the widget.

        Returns:
            The Jupyter widget container.
        """
        self._create_ui()
        display(self._container)

        self._init_plot()
        self._running = True
        self._status_label.value = "<b>Status:</b> Streaming..."

        # Start streaming in background thread
        self._stream_thread = threading.Thread(target=self._stream_data)
        self._stream_thread.start()

        # Start update loop
        self._update_loop()

        return self._container

    def _update_loop(self) -> None:
        """Periodic update loop."""
        if self._running:
            self._update_plot()
            # Schedule next update
            loop = asyncio.get_event_loop()
            loop.call_later(self._update_interval, self._update_loop)

    def stop(self) -> None:
        """Stop streaming."""
        self._running = False
        self._client.stop_simulation(self._session_id)
        self._status_label.value = "<b>Status:</b> Stopped"
        self._progress_bar.bar_style = "warning"

    def get_data(self) -> Tuple[np.ndarray, Dict[str, np.ndarray]]:
        """Get the collected data.

        Returns:
            Tuple of (times, {signal_name: values}).
        """
        times = np.array(self._times)
        signals = {name: np.array(vals) for name, vals in self._values.items()}
        return times, signals


class InteractivePlot:
    """Plotly-based interactive streaming plot for Jupyter notebooks.

    Provides a more interactive experience with zoom, pan, and hover.

    Example:
        >>> from pulsim.grpc.widgets import InteractivePlot
        >>>
        >>> plot = InteractivePlot(client, session_id, signals=["V(out)"])
        >>> plot.start()
    """

    def __init__(
        self,
        client: "PulsimClient",
        session_id: str,
        signals: Optional[List[str]] = None,
        decimation: int = 1,
        max_points: int = 10000,
        update_interval: float = 0.2,
        height: int = 500,
    ):
        """Initialize the interactive plot.

        Args:
            client: Connected PulsimClient instance.
            session_id: Session to stream from.
            signals: Signals to plot.
            decimation: Decimation factor.
            max_points: Maximum points to display.
            update_interval: Update interval in seconds.
            height: Plot height in pixels.
        """
        _check_dependencies(plotly=True)

        self._client = client
        self._session_id = session_id
        self._signals = signals
        self._decimation = decimation
        self._max_points = max_points
        self._update_interval = update_interval
        self._height = height

        self._times: List[float] = []
        self._values: Dict[str, List[float]] = {}
        self._signal_names: List[str] = []
        self._running = False
        self._stream = None
        self._fig_widget = None
        self._output = None

    def _create_ui(self) -> None:
        """Create the UI components."""
        # Status
        self._status_label = widgets.HTML(value="<b>Status:</b> Initializing...")

        # Control buttons
        self._stop_button = widgets.Button(
            description="Stop",
            button_style="danger",
            icon="stop",
        )
        self._stop_button.on_click(lambda _: self.stop())

        controls = widgets.HBox([self._stop_button])

        # Figure widget
        self._fig = go.FigureWidget()
        self._fig.update_layout(
            height=self._height,
            xaxis_title="Time (s)",
            yaxis_title="Value",
            showlegend=True,
            legend=dict(yanchor="top", y=0.99, xanchor="right", x=0.99),
        )

        self._container = widgets.VBox([
            self._status_label,
            controls,
            self._fig,
        ])

    def _stream_data(self) -> None:
        """Stream data from the server."""
        try:
            self._stream = self._client.stream_waveforms(
                self._session_id,
                self._signals,
                self._decimation,
            )

            for sample in self._stream:
                if not self._running:
                    break

                self._times.append(sample.time)
                for sig_name, value in sample.values.items():
                    if sig_name not in self._signal_names:
                        self._signal_names.append(sig_name)
                        self._values[sig_name] = []
                        # Add trace
                        self._fig.add_trace(
                            go.Scatter(x=[], y=[], name=sig_name, mode="lines")
                        )
                    self._values[sig_name].append(value)

            self._status_label.value = "<b>Status:</b> Completed"

        except Exception as e:
            self._status_label.value = f"<b>Status:</b> Error - {e}"

        finally:
            self._running = False

    def _update_plot(self) -> None:
        """Update the plot with new data."""
        if not self._times or not self._fig.data:
            return

        times = self._times[-self._max_points:]

        with self._fig.batch_update():
            for i, sig_name in enumerate(self._signal_names):
                if i < len(self._fig.data):
                    values = self._values.get(sig_name, [])[-self._max_points:]
                    self._fig.data[i].x = times
                    self._fig.data[i].y = values

    def start(self) -> widgets.VBox:
        """Start streaming and display."""
        self._create_ui()
        display(self._container)

        self._running = True
        self._status_label.value = "<b>Status:</b> Streaming..."

        # Start streaming in background
        self._stream_thread = threading.Thread(target=self._stream_data)
        self._stream_thread.start()

        # Start update loop
        self._update_loop()

        return self._container

    def _update_loop(self) -> None:
        """Periodic update."""
        if self._running:
            self._update_plot()
            loop = asyncio.get_event_loop()
            loop.call_later(self._update_interval, self._update_loop)

    def stop(self) -> None:
        """Stop streaming."""
        self._running = False
        self._client.stop_simulation(self._session_id)
        self._status_label.value = "<b>Status:</b> Stopped"

    def get_data(self) -> Tuple[np.ndarray, Dict[str, np.ndarray]]:
        """Get collected data."""
        times = np.array(self._times)
        signals = {name: np.array(vals) for name, vals in self._values.items()}
        return times, signals


def plot_waveforms(
    client: "PulsimClient",
    session_id: str,
    signals: Optional[List[str]] = None,
    interactive: bool = True,
    **kwargs,
) -> StreamingPlot:
    """Convenience function to create a streaming plot.

    Args:
        client: PulsimClient instance.
        session_id: Session identifier.
        signals: Signals to plot.
        interactive: Use Plotly (True) or Matplotlib (False).
        **kwargs: Additional arguments for the plot class.

    Returns:
        The plot widget (started).
    """
    if interactive and _HAS_PLOTLY:
        plot = InteractivePlot(client, session_id, signals, **kwargs)
    else:
        plot = StreamingPlot(client, session_id, signals, **kwargs)

    return plot.start()
